function createRandomIdentifier(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

//
// credit: https://stackoverflow.com/a/51086910/2005038
// by user Bergi https://stackoverflow.com/users/1048572/bergi
//
const MutexImplementationFromStackOverflow = class {
	constructor() {
		this._lock = null;
	}
	isLocked() {
		return this._lock != null;
	}
	_acquire() {
		var release;
		const lock = this._lock = new Promise(resolve => {
			release = resolve;
		});
		return () => {
			if (this._lock == lock) this._lock = null;
			release();
		};
	}
	acquireSync() {
		if (this.isLocked()) throw new Error("still locked!");
		return this._acquire();
	}
	acquireQueued() {
		const q = Promise.resolve(this._lock).then(() => release);
		const release = this._acquire(); // reserves the lock already, but it doesn't count
		return q; // as acquired until the caller gets access to `release` through `q`
	}
};

function createAsyncMutex() {
	let mutex = new MutexImplementationFromStackOverflow();

	return {
		acquire() {
			return mutex.acquireQueued()
		}
	}
}

function sendSingleShotRequestWithTimeout(instance, data, timeout = 0) {
	instance.assertReadyAndNotClosed();

	// create a request id that identifies this particular request
	const request_id = createRandomIdentifier(12);
	// create the promise returned by sendSingleShotRequest()
	const request_promise = createPromise();
	// keep track of timer to clear the timeout when response arrives
	let timeout_timer = null;

	if (timeout > 0) {
		timeout_timer = setTimeout(() => {
			const {reject} = request_promise;

			setTimeout(reject, 0, new Error(`Timeout of '${timeout}ms' reached.`));

			instance.open_requests.delete(request_id);
		}, timeout);
	}

	instance.open_requests.set(request_id, {
		request_promise,
		timer: timeout_timer
	});

	instance.sendJSONData({
		cmd: "request",
		request_id,
		data
	});

	return request_promise.promise
}

function sendRetransmittingRequest(instance, data, max_attempts = -1) {
	instance.assertReadyAndNotClosed();

	// create a request id that identifies this particular request
	const request_id = createRandomIdentifier(12);

	instance.trace(`i will be using requestId '${request_id}' for the message '${JSON.stringify(data)}'.`);

	// create the promise returned by sendRequest()
	const request_promise = createPromise();
	// save retransmission delay for this request
	const retransmission_delay = instance.retransmission_delay;

	let attempt_no = 0;

	const retransmit = () => {
		if (!instance.open_requests.has(request_id)) {
			// no need for retransmit, request does not exist anymore
			// this happens when the request is either fullfilled or
			// close() was called.

			instance.debug(`no need for retransmitting request '${request_id}'`);

			return
		}

		const open_request = instance.open_requests.get(request_id);

		++attempt_no;

		if (attempt_no > max_attempts && max_attempts >= 0) {
			const error_msg = `Max retransmission attempts of ${max_attempts} reached.`;

			instance.debug(error_msg);

			setTimeout(request_promise.reject, 0, new Error(error_msg));

			instance.open_requests.delete(request_id);

			return
		}

		instance.debug(`i'm retransmitting the request '${request_id}'`);

		instance.sendJSONData({
			cmd: "request",
			request_id,
			data,
			max_attempts,
			attempt_no
		});

		open_request.timer = setTimeout(retransmit, retransmission_delay);
	};

	instance.open_requests.set(request_id, {
		request_promise,
		timer: setTimeout(retransmit, retransmission_delay)
	});

	instance.sendJSONData({
		cmd: "request",
		request_id,
		data,
		max_attempts,
		attempt_no
	});

	return request_promise.promise
}

function pruneJavaScriptMapObject(map, max_size = 1000) {
	if (max_size > map.size) {
		return
	}

	const map_keys = Array.from(map.keys());
	const keys_to_be_removed = map_keys.slice(
		0, map.size - max_size
	);

	for (const key of keys_to_be_removed) {
		map.delete(key);
	}
}

function pruneRequestsCache(instance) {
	pruneJavaScriptMapObject(instance.handled_requests);
	pruneJavaScriptMapObject(instance.received_requests);
}

async function handleMessage(instance, message) {
	instance.trace(`i will be handling the following message '${JSON.stringify(message)}'`);

	//
	// call request handler
	//
	if (message.cmd === "request") {
		let response = null, from_cache = false;

		if (instance.handled_requests.has(message.request_id)) {
			response = instance.handled_requests.get(message.request_id);

			from_cache = true;
		} else {
			if (!instance.received_requests.has(message.request_id)) {

				if (!instance.debug_options.disable_saving_received_requests) {
					instance.received_requests.set(message.request_id, 1);
				}

				response = await instance.public_interface.requestHandler(message.data, null, {
					debug: {
						instance,
						message
					}
				});
			} else {
				instance.debug(`i already received this request: '${message.request_id}'`);

				return
			}
		}

		let from_cache_str = from_cache ? " (from cache)" : "";

		instance.trace(`i have already handled the message '${JSON.stringify(message)}'. my response is '${JSON.stringify(response)}'${from_cache_str}`);

		instance.sendJSONData({
			cmd: "response",
			original_request_id: message.request_id,
			response,
			from_cache
		});

		if (!from_cache) {
			instance.handled_requests.set(message.request_id, response);
		}

		pruneRequestsCache(instance);
	}
	//
	// handle incoming response
	//
	else if (message.cmd === "response") {
		const {original_request_id} = message;

		if (!instance.open_requests.has(message.original_request_id)) {
			instance.debug(`no pending request with id '${original_request_id}'`);

			return
		}

		const open_request = instance.open_requests.get(original_request_id);

		if (open_request.timer !== null) {
			clearTimeout(open_request.timer);
		}

		const {resolve} = open_request.request_promise;

		setTimeout(resolve, 0, message.response);

		instance.open_requests.delete(original_request_id);
	}
}

async function onMessageReceived(instance, message) {
	//
	// make sure every request is handled sequentially
	// this is needed for the response cache
	//
	let release = () => {};

	if (!instance.debug_options.disable_mutex) {
		release = await instance.mutex.acquire();
	}

	try {
		await handleMessage(instance, message);
	} finally {
		await release();
	}
}

function createRequestResponseProtocol(api, label = "") {
	const connection_id = createRandomIdentifier(12);
	const synchronize_token = createRandomIdentifier(32);
	let synchronized_promise = createPromise();

	/*
	 * Wrap api.sendMessage to ignore errors that are thrown
	 * by that function.
	 */
	const sendMessageViaAPI = async (instance, message) => {
		try {
			await api.sendMessage(message);
		} catch (e) {
			instance.debug(`api.sendMessage failed with '${e.message}'`);
		}
	};

	let instance = {
		connection_id,

		open_requests: new Map(),
		mutex: createAsyncMutex(),
		/*
		 * Keep a map of handled requests so if the same
		 * request is received a second time, the result is re-used
		 * and the request handler isn't called a second time.
		 */
		handled_requests: new Map(),
		/**
		 * Keep a map of received requests to not execute the
		 * requestHandler a second time. Yes, this can happen.
		 */
		received_requests: new Map(),
		ready: false,
		closed: false,

		retransmission_delay: 350,

		debug(...args) {
			if (instance.public_interface.debug !== true) return

			console.log(`[${label} (D)]`, ...args);
		},

		// only to be used for debugging
		debug_options: {
			disable_mutex: false,
			disable_saving_received_requests: false
		},

		trace(...args) {
			if (instance.public_interface.trace !== true) return

			console.log(`[${label} (T)]`, ...args);
		},

		sendJSONData(data) {
			sendMessageViaAPI(instance, `start{` + JSON.stringify(data) + `}end`);
		},

		messageHandler(message) {
			if (message.startsWith(`@anio-js-foundation/requestResponseProtocol:`)) {
				message = message.slice(`@anio-js-foundation/requestResponseProtocol:`.length);

				if (message.startsWith("sync:")) {
					const sync_token = message.slice("sync:".length);

					sendMessageViaAPI(instance, `@anio-js-foundation/requestResponseProtocol:ack:${sync_token}`);
				} else if (message.startsWith("ack:")) {
					const sync_token = message.slice("ack:".length);

					if (sync_token === synchronize_token && !instance.ready) {
						instance.ready = true;

						synchronized_promise.resolve();
					}
				} else {
					instance.debug("bad message", message);
				}
			} else if (instance.ready && message.startsWith("start{") && message.endsWith("}end")) {
				message = message.slice("start{".length);
				message = message.slice(0, message.length - "}end".length);

				onMessageReceived(instance, JSON.parse(message));
			} else {
				instance.debug("bad message", message);
			}
		},

		assertReadyAndNotClosed() {
			if (!instance.ready) {
				throw new Error(`Cannot send request when protocol is not ready.`)
			} else if (instance.closed) {
				throw new Error(`Cannot send request when protocol is closed.`)
			}
		},

		public_interface: {
			debug: false,
			trace: false,

			connection_id,

			_setDebugDisableMutex(value) {
				instance.debug_options.disable_mutex = value;
			},

			_setDisableSavingReceivedRequests(value) {
				instance.debug_options.disable_saving_received_requests = value;
			},

			ready() {
				return synchronized_promise.promise
			},

			setRetransmissionDelay(amount) {
				instance.retransmission_delay = amount;
			},

			requestHandler(request) {
				return `Please set instance.requestHandler to handle requests. Incoming request data was: ${JSON.stringify(request)}.`
			},

			withTimeout(timeout_value) {
				return {
					sendRequest(request_data) {
						let max_attempts = Math.floor(timeout_value / instance.retransmission_delay);

						instance.debug(
							`Calculated max_attempts = '${max_attempts}' from timeout value '${timeout_value}ms' (retransmission delay is '${instance.retransmission_delay}ms')`
						);

						return sendRetransmittingRequest(instance, request_data, max_attempts)
					},

					sendSingleShotRequest(request_data) {
						return sendSingleShotRequestWithTimeout(instance, request_data, timeout_value)
					}
				}
			},

			withMaxRetransmissionAttempts(max_attempts) {
				return {
					sendRequest(request_data) {
						return sendRetransmittingRequest(instance, request_data, max_attempts)
					}
				}
			},

			sendRequest(request_data) {
				return sendRetransmittingRequest(instance, request_data, -1)
			},

			sendSingleShotRequest(request_data) {
				return sendSingleShotRequestWithTimeout(instance, request_data, 0)
			},

			closeAllPendingRequests(reason = `closeAllPendingRequests() was called.`) {
				for (const [request_id, open_request] of instance.open_requests) {
					const {reject} = open_request.request_promise;

					setTimeout(reject, 0, new Error(reason));

					// make sure to cancel timer
					if (open_request.timer !== null) {
						clearTimeout(open_request.timer);
					}

					instance.open_requests.delete(request_id);
				}
			},

			close() {
				if (instance.closed) return

				instance.closed = true;

				setTimeout(
					instance.public_interface.closeAllPendingRequests, 0, "close() was called."
				);

				api.removeEventListener("message", instance.messageHandler);
			}
		}
	};

	api.on("message", instance.messageHandler);

	/**
	 * Slave can miss this message so repeat it
	 * every once in a while to get the master and slave
	 * synced up.
	 */
	const delay_map = [50, 50, 50, 100, 100, 150, 150, 150, 250, 500, 750, 1000];
	let delay_index = 0;

	const synchronize = async () => {
		if (instance.closed) return
		if (instance.ready) return

		instance.debug("attempt to synchronize");

		await sendMessageViaAPI(instance, `@anio-js-foundation/requestResponseProtocol:sync:${synchronize_token}`);

		// by default use the last value from the delay map
		let amount = delay_map[delay_map.length - 1];

		if (delay_map.length > delay_index) {
			amount = delay_map[delay_index];

			++delay_index;
		}

		instance.debug(`will be calling synchronize() in ${amount}ms`);

		setTimeout(synchronize, amount);
	};

	setTimeout(synchronize, 0);

	return instance.public_interface
}

export { createRequestResponseProtocol as default };
